﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Wecker                                                                              //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright © 2016 Reimar Grasbon (softwork@grasbon.de)                               //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "Timer.h"
#include "String.h"
#include "Display.h"
#include "Board.h"
#include "Sound.h"
#include "Storing.h"
#include "Application.h"
#include "Alarm.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                  // Zustände des Weckers:
{   AS_OFF     = 0,           // · Wecker ausgeschaltet
    AS_SET     = 1,           // · Wecker gestellt
    AS_BLOCKED = 2,           // · Wecker blockiert
    AS_ALARM   = 3,           // · Weckalarm
    AS_SNOOZE  = 4,           // · Schlummern
    AS_QUIT    = 5            // · Wecken beendet
} AlarmState;

typedef enum                  // Aktivierung des Weckers:
{   AE_ENABLE,                // · Wecker einschalten
    AE_DISABLE,               // · Wecker ausschalten
    AE_NONE                   // · Keine Aktion
} AlarmEnabler;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static AlarmState    state       = AS_OFF;           // Zustand des Weckautomaten
static AlarmEnabler  enabler     = AE_NONE;          // Auslöser zum Ein-/Ausschalten
static DateTime      nextAlarmDt = SILVESTER_2000;   // Nächster Wecktermin in UTC

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen   //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  ForerunPhase(): Die Funktion liefert 'true', wenn die angegebene Vorlaufzeitphase   //
//  --------------  aktiv ist. Die Vorlaufzeitphase ist ein Zeitabschnitt vor Eintritt  //
//  des nächsten Wecktermins. Die Dauer der Vorlaufzeitphase wird in 'length' fest-     //
//  gelegt. Ist der Wecker ausgeschaltet, kehrt die Funktion mit 'false' zurück.        //
//                                                                                      //
//  Eingabe: · length: Länge der Vorlaufzeitphase: 00:00:00 - 23:59:59.                 //
//  Ausgabe: 'true', wenn die Vorlaufzeitphase aktiv ist.							    //
//////////////////////////////////////////////////////////////////////////////////////////
static bool
ForerunPhase(const Time* length)
{
    if( state == AS_OFF ) return false;
    DateTime  dateTime;
    BOARD_Clock(&dateTime);
    if( !DATETIME_IsValid(&dateTime) ) return false;
    DATETIME_ShiftHour(&dateTime,   length->hour);
    DATETIME_ShiftMinute(&dateTime, length->minute);
    DATETIME_ShiftSecond(&dateTime, length->second);
    return DATETIME_Compare(&dateTime, &nextAlarmDt) >= 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  AlarmPhase(): Die Funktion liefert 'true', wenn die Weckzeitphase aktiv ist. Das    //
//  ------------  ist dann der Fall, wenn der Wecker eingeschaltet ist und die aktuelle //
//  Uhrzeit den Wecktermin erreicht oder überschritten hat.                             //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn die Weckzeitphase aktiv ist.									//
//////////////////////////////////////////////////////////////////////////////////////////
static bool
AlarmPhase(void)
{
    Time phase = { 0,0,0 };           // Vorlaufphase ohne Vorlaufzeit = Wecktermin
    return ForerunPhase(&phase);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ReactivationPhase(): Die Funktion liefert 'true', wenn der Wecker sich in der Re-   //
//  -------------------  aktivierungsphase befindet. Die Reaktivierungsphase ist eine   //
//  Zeitspanne vor Eintritt des nächsten Wecktermins, zu deren Beginn der Wecker nach   //
//  einem abgelaufenen Weckalarm wieder gestellt wird (dazu siehe auch ALARM_IsSet()).  //
//  Die Reaktivierungsphase beginnt 'CFG_REARM_ALARM_LEAD_TIME' Stunden vor Eintritt    //
//  des nächsten Wecktermins. Ist der Wecker ausgeschaltet, kehrt die Funktion mit      //
//  'false' zurück.                                                                     //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn die Reaktivierungsphase aktiv ist.							//
//////////////////////////////////////////////////////////////////////////////////////////
static bool
ReactivationPhase(void)
{
    Time phase = { CFG_REARM_ALARM_LEAD_TIME,0,0 };
    return ForerunPhase(&phase);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  SetNextAlarmDateTime(): Ermittelt auf Basis der aktuellen Zeitinformation und der   //
//  ----------------------  ausgewählten Weckbetriebsart den Zeitpunkt des nächsten     //
//  Wecktermins. Der ermittelte Wecktermin wird als UTC-Zeit in 'nextAlarmDt' abgelegt. //
//                                                                                      //
//  Eingabe: · alarmOperation: Die aktuelle Weckbetriebsart gemäß 'AlarmOperation'.     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetNextAlarmDateTime(AlarmOperation alarmOperation)
{
    // (1) Nächsten Wecktermin gemäß AO_DAILY oder AO_ONCE_ONLY bestimmen
    DateTime dateTime;
    BOARD_Clock(&dateTime);                     // Aktuelle Datumzeit in UTC
    if( !DATETIME_IsValid(&dateTime) ) return;
    bool autoDstEnabled;
    STORING_Load(STO_AutoDstEnabled, &autoDstEnabled);
    DATETIME_Localize(CFG_TIMEZONE, autoDstEnabled, &dateTime);
    Time actualTime = dateTime.time;			// Aktuelle lokale Zeit
    Time alarmTime;
    STORING_Load(STO_AlarmTime, &alarmTime);    // Weckzeit in lokaler Zeit		
    if( !TIME_IsValid(&alarmTime) ) return;
    dateTime.time = alarmTime;					// Weckzeit mit Datum von heute (lokal)
    if( TIME_Compare(&alarmTime, &actualTime) <= 0 ) DATE_ShiftDay(&dateTime.date, +1);	
    // (2) Abhängig von der Weckbetriebsart, Wecktermin weiterschieben
    bool shiftDay;
    while( true )
    {   DTWeekday weekday   = DATE_WeekDay(&dateTime.date);
        bool      isWeekEnd = weekday == DT_SAT || weekday == DT_SUN;
        bool      isHoliday = APPLICATION_IsHoliday(DATE_DayType(&dateTime.date));
        switch( alarmOperation )
        {   case AO_DAILY_EX_WEEKEND: shiftDay = isWeekEnd;                break;
            case AO_WORKDAY:          shiftDay = isWeekEnd || isHoliday;   break;
            case AO_WEEKEND_ONLY:     shiftDay = !isWeekEnd;               break;
            case AO_WEEKEND_HOLIDAY:  shiftDay = !isWeekEnd && !isHoliday; break;
            case AO_HOLIDAY_ONLY:     shiftDay = !isHoliday;               break;
            default:                  shiftDay = false;      // AO_DAILY oder AO_ONCE_ONLY
        }
        if( !shiftDay ) break;
        DATETIME_ShiftDay(&dateTime, +1);
    };
    DATETIME_Universalize(CFG_TIMEZONE, autoDstEnabled, &dateTime);
    nextAlarmDt = dateTime;	// Nächster Wecktermin in UTC
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  ALARM_Enable(): Schaltet den Wecker ein oder aus.                                   //
//  --------------                                                                      //
//                                                                                      //
//  Eingabe: · enable: 'true'/'false', Wecker ein-/ausschalten.                         //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
ALARM_Enable(bool enable)
{
    if( enable ) enabler = AE_ENABLE;
    else         enabler = AE_DISABLE;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ALARM_IsEnabled(): Ermittelt, ob der Wecker eingeschaltet ist.                      //
//  -----------------                                                                   //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true' wenn der Wecker eingeschaltet ist.									//
//////////////////////////////////////////////////////////////////////////////////////////
bool
ALARM_IsEnabled(void)
{
    return state != AS_OFF || enabler == AE_ENABLE;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ALARM_IsSet(): Prüft, ob der Wecker gestellt ist. Der Wecker gilt als gestellt,     //
//  -------------  wenn er eingeschaltet ist und sich in der Reaktivierungsphase be-    //
//  findet (siehe auch ReactivationPhase() ).                                           //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true' wenn der Wecker gestellt ist.                                       //
//////////////////////////////////////////////////////////////////////////////////////////
bool
ALARM_IsSet(void)
{
    return state == AS_SET || state == AS_BLOCKED;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ALARM_IsAlarm(): Prüft, ob Weckalarm vorliegt. Auch der Schlummerbetrieb gilt als   //
//  ---------------  Weckalarm.                                                         //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn Weckalarm oder Schlummerbetrieb aktiv ist.                    //
//////////////////////////////////////////////////////////////////////////////////////////
bool
ALARM_IsAlarm(void)
{
	return state == AS_ALARM || state == AS_SNOOZE;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ALARM_IsSnooze(): Prüft, ob gerade Schlummerbetrieb vorliegt.                       //
//  ----------------                                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn Schlummerbetrieb aktiv ist.									//
//////////////////////////////////////////////////////////////////////////////////////////
bool
ALARM_IsSnooze(void)
{
    return state == AS_SNOOZE;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ALARM_AnnouncementPhase(): Die Funktion liefert 'true', wenn die Weckankündigungs-  //
//  -------------------------  phase aktiv ist. Die Weckankündigungsphase ist der Zeit- //
//  abschnitt, in dem der Wecker einen bevorstehenden Weckalarm ankündigt. Die Weckan-  //
//  kündigungsphase beginnt 'CFG_ALARM_ANNOUNCEMENT' Sekunden vor Eintritt des nächsten //
//  Wecktermins. Ist der Wecker ausgeschaltet, kehrt die Funktion mit 'false' zurück.   //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn die Weckankündigungsphase aktiv ist.							//
//////////////////////////////////////////////////////////////////////////////////////////
bool
ALARM_AnnouncementPhase(void)
{
    Time phase = { 0,0,CFG_ALARM_ANNOUNCEMENT };
    return ForerunPhase(&phase);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ALARM_Proceed(): Verfahrensfunktion des Weckautomaten. Der Automat hat folgendes    //
//  ---------------  Zustandsdiagramm:                                                  //
//                                                                                      //
//                                                                                      //
//                          AON,ENABLE()                                 AON,ENABLE()   //
//                          +--------+                                   +--------+     //
//                          |       \|/                                  |       \|/    //
//                      +---+------------+                             +-+----------+   //
//      REA,REARM()     |////////////////| CLR,NONE()                  |////////////|   //
//      +-------------->|////////////////+---------------------------->|////////////|   //
//      |  AON,ENABLE() |////////////////|                             |////////////|   //
//      |  +----------->|// AS_BLOCKED //|                  BLK,NONE() |////////////|   //
//      |  |            |////////////////|<----------------------------+////////////|   //
//      |  |  +---------+////////////////|<---+                        |////////////|   //
//      |  |  |         |////////////////|    |                        |////////////|   //
//      |  |  |         +--------+-------+    |                        |// AS_SET //|   //
//      |  |  |    AOF,DISABLE() |            | AON,ENABLE()           |////////////|   //
//      |  |  |                  |     +------+------+                 |////////////|   //
//      |  |  |                  |     |/////////////|   AOF,DISABLE() |////////////|   //
//      |  |  |                  +---->|// AS_OFF ///|<----------------+////////////|   //
//      |  | ALM,PAUSE()               |/////////////|                 |////////////|   //
//      |  |  |                        +-------------+                 +--------+---+   //
//      |  |  |                         /|\  /|\  /|\               ALM,ALARM() |       //
//      |  |  |                          |    |    |                            |       //
//      |  |  |    AOF,DISABLE()         |    |    |           AOF,DISABLE()    |       //
//      |  |  |   +----------------------+    |    +------------------------+   |       //
//      |  |  |   |                           |                             |   |       //
//      |  |  |   |                           | AOF,DISABLE()               |   |       //
//      |  |  |   |                 +---------+-----+                       |   |       //
//      |  |  |   |                 |///////////////+-------------------+   |   |       //
//      |  |  |   |                 |///////////////| ZAC,ALARM()       |   |   |       //
//      |  | \|/  |                 |///////////////|                  \|/  |  \|/      //
//    +-+--+------+-+               |// AS_SNOOZE //|                +------+-------+   //
//    |/////////////|   STP,PAUSE() |///////////////|   ZZZ,SNOOZE() |//////////////|   //
//    |/////////////|<--------------|///////////////|<---------------+//////////////|   //
//    |/////////////|               |///////////////|                |//////////////|   //
//    |// AS_QUIT //|               +---------------+    TME,PAUSE() |// AS_ALARM //|   //
//    |/////////////|<-----------------------------------------------+//////////////|   //
//    |/////////////|                                    STP,PAUSE() |//////////////|   //
//    |/////////////|<-----------------------------------------------+//////////////|   //
//    +-------------+                                                +--------------+   //
//                                                                                      //
//                                                                                      //
//  Zustände:                                  Ereignisse (in absteigender Priorität):  //
//   · AS_OFF:     Wecker ausgeschaltet        · AON: Wecker einschalten                //
//   · AS_SET:     Wecker gestellt             · AOF: Wecker abschalten                 //
//   · AS_BLOCKED: Wecker blockiert            · ALM: Weckzeit erreicht                 //
//   · AS_ALARM:   Weckalarm                   · REA: Reaktivierungsphase erreicht      //
//   · AS_SNOOZE:  Schlummerbetrieb            · STP: Weckalarm abgeschaltet            //
//   · AS_QUIT:    Wecken beendet              · TME: Weckalarm-Zeit abgelaufen         //
//                                             · ZZZ: Schlummerbetrieb aktiviert        //
//  Aktionen:                                  · ZAC: Schlummerbetriebszeit abgelaufen  //
//   · ALARM():   Weckalarm starten            · BLK: Wecken blockieren                 //
//   · PAUSE():   Weckalarm beenden            · CLR: Wecken erlauben                   //
//   · SNOOZE():  Schlummerbetrieb einleiten                                            //
//   · REARM():   Wecker reaktivieren                                                   //
//   · ENABLE():  Wecker einschalten                                                    //
//   · DISABLE(): Wecker abschalten                                                     //
//   · NONE():    Keine Aktion                                                          //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
void
ALARM_Proceed(void)
{
    enum   // Ereignisse (in absteigender Priorität)
    {   AON = BV(0),
        AOF = BV(1),
        ALM = BV(2),
        REA = BV(3),
        STP = BV(4),
        TME = BV(5),
        ZZZ = BV(6),
        ZAC = BV(7),
        BLK = BV(8),
        CLR = BV(9)
    };

    typedef enum
    {   ALARM, PAUSE, SNOOZE, REARM, ENABLE, DISABLE, NONE
    } Action;

    typedef struct
    {   AlarmState  state;
        Action      action;
    } Entry;

    static const Entry zeta[][6] PROGMEM =
    {   //+------ AS_OFF ------+,+------ AS_SET -------+,+---- AS_BLOCKED -----+,+----- AS_ALARM -----+,+---- AS_SNOOZE -----+,+------ AS_QUIT ------+
        { { AS_BLOCKED, ENABLE },{ AS_SET    , ENABLE  },{ AS_BLOCKED, ENABLE  },{ AS_ALARM , NONE    },{ AS_SNOOZE, NONE    },{ AS_BLOCKED, ENABLE  } }, // AON
        { { AS_OFF    , NONE   },{ AS_OFF    , DISABLE },{ AS_OFF    , DISABLE },{ AS_OFF   , DISABLE },{ AS_OFF   , DISABLE },{ AS_OFF    , DISABLE } }, // AOF
        { { AS_OFF    , NONE   },{ AS_ALARM  , ALARM   },{ AS_QUIT   , PAUSE   },{ AS_ALARM , NONE    },{ AS_SNOOZE, NONE    },{ AS_QUIT   , NONE    } }, // ALM
        { { AS_OFF    , NONE   },{ AS_SET    , NONE    },{ AS_BLOCKED, NONE    },{ AS_ALARM , NONE    },{ AS_SNOOZE, NONE    },{ AS_BLOCKED, REARM   } }, // REA
        { { AS_OFF    , NONE   },{ AS_SET    , NONE    },{ AS_BLOCKED, NONE    },{ AS_QUIT  , PAUSE   },{ AS_QUIT  , PAUSE   },{ AS_QUIT   , NONE    } }, // STP
        { { AS_OFF    , NONE   },{ AS_SET    , NONE    },{ AS_BLOCKED, NONE    },{ AS_QUIT  , PAUSE   },{ AS_SNOOZE, NONE    },{ AS_QUIT   , NONE    } }, // TME
        { { AS_OFF    , NONE   },{ AS_SET    , NONE    },{ AS_BLOCKED, NONE    },{ AS_SNOOZE, SNOOZE  },{ AS_SNOOZE, NONE    },{ AS_QUIT   , NONE    } }, // ZZZ
        { { AS_OFF    , NONE   },{ AS_SET    , NONE    },{ AS_BLOCKED, NONE    },{ AS_ALARM , NONE    },{ AS_ALARM , ALARM   },{ AS_QUIT   , NONE    } }, // ZAC
        { { AS_OFF    , NONE   },{ AS_BLOCKED, NONE    },{ AS_BLOCKED, NONE    },{ AS_ALARM , NONE    },{ AS_SNOOZE, NONE    },{ AS_QUIT   , NONE    } }, // BLK
        { { AS_OFF    , NONE   },{ AS_SET    , NONE    },{ AS_SET    , NONE    },{ AS_ALARM , NONE    },{ AS_SNOOZE, NONE    },{ AS_QUIT   , NONE    } }, // CLR
    };

    static Timer  alarmTimer  = TIMER(TIMER_SECONDS, 0);
    static Timer  snoozeTimer = TIMER(TIMER_SECONDS, 0);

    bool alarmByRadio;
    STORING_Load(STO_AlarmTypeRadio, &alarmByRadio);
    // Ereigniserzeugung
    UInt16  eventSet = 0;
    // AON: Wecker explizit einschalten
    if( enabler == AE_ENABLE ) eventSet |= AON;
    // AOF: Wecker explizit abschalten
    if( enabler == AE_DISABLE ) eventSet |= AOF;
    enabler = AE_NONE;
    // ALM: Weckzeit erreicht
    if( AlarmPhase() ) eventSet |= ALM;
    // REA: Reaktivierungszeit erreicht
    if( ReactivationPhase() ) eventSet |= REA;
    // STP: Weckalarm abgeschaltet
    if( state == AS_ALARM || state == AS_SNOOZE )
    {   if( APPLICATION_UserInteraction() ) eventSet |= STP;
    }
    // TME: Weckalarm-Zeit abgelaufen
    if( TIMER_HasExpired(&alarmTimer) ) eventSet |= TME;
    // ZZZ: Schlummerbetrieb aktiviert
    {   bool  enabled;
        STORING_Load(STO_SnoozeEnabled, &enabled);
        if( enabled )
        {   static Timer update = TIMER(TIMER_MILLISECONDS, 50);
            if( TIMER_RestartIfExpired(&update) )
            {   static SInt08 sensitivity = 0;
                if( BOARD_PIRSensor() )
                {   SInt08 sensitivityLevel;
                    STORING_Load(STO_MotionSensDisplay, &sensitivityLevel);
                    sensitivityLevel = 9 - sensitivityLevel;
                    if( sensitivity < sensitivityLevel*8 ) sensitivity++;
                    else                                   eventSet |= ZZZ;
                }
                else
                {   if( sensitivity > 0 ) sensitivity--;
                }
            }
        }
    }
    // ZAC: Schlummerbetriebszeit abgelaufen
    if( TIMER_HasExpired(&snoozeTimer) ) eventSet |= ZAC;
    // BLK: Wecken blockieren
    if( APPLICATION_Mode() != OM_CLOCK ) eventSet |= BLK;
    if( alarmByRadio && APPLICATION_RadioManuallyOn() ) eventSet |= BLK;
    // CLR: Wecken erlauben
    if( APPLICATION_Mode() == OM_CLOCK )
    {   if( !alarmByRadio || !APPLICATION_RadioManuallyOn() ) eventSet |= CLR;
    }
    // Zustandsübergang
    Action action = NONE;
    for( UInt08 i=0; i<CARDINALITY(zeta); i++ )
    {   const Entry* entry = &zeta[i][(UInt08)state];
        if( (eventSet & BV(i)) != 0 )
        {   AlarmState newState = (AlarmState)(pgm_read_byte(&entry->state));
            Action     tmpAction = (Action)(pgm_read_byte(&entry->action));
            if( state == newState && tmpAction == NONE ) continue;
            state  = newState;
            action = tmpAction;
            break;
        }
    }
    // Aktionsausführung
    switch( action )
    {   case ALARM:
        {   if( alarmByRadio )
            {   BOARD_EnableRelais(true);
                SOUND_Start(SOUND_ALARM_BEGIN);
            }
            else
            {   SOUND_Start(SOUND_ALARM_SIGNAL);
            }
            UInt08 minutes;
            STORING_Load(STO_AlarmDuration, &minutes);
            TIMER_Start(&alarmTimer, (SInt16)minutes*60);
            DISPLAY_StartAnim(ANIM_ALARM);
            DISPLAY_SetBlinkingDigits(0b00000011);
            break;
        }
        case PAUSE:
        {   if( alarmByRadio && !APPLICATION_RadioManuallyOn() )
            {   BOARD_EnableRelais(false);
                BOARD_SetLedBrightness(0);
            }
            DISPLAY_StartAnim(ANIM_ALARM_PAUSE);
            DISPLAY_SetBlinkingDigits(0);
            SOUND_Start(SOUND_ALARM_PAUSE);
            AlarmOperation alarmOperation;
            STORING_Load(STO_AlarmOperation, &alarmOperation);
            SetNextAlarmDateTime(alarmOperation);
            break;
        }
        case SNOOZE:
        {   if( alarmByRadio )
            {   BOARD_EnableRelais(false);
                BOARD_SetLedBrightness(0);              // Wird die Schlummerfunktion in
            }                                           // den ersten 20 Sekunden der
            UInt08 minutes;                             // laufenden Minute aktiviert,
            STORING_Load(STO_SnoozeDuration, &minutes); // wird die Dauer des Schlummer-
            SInt16 seconds = (SInt16)minutes*60;        // betriebs auf ganze Minuten
            DateTime  dateTime;                         // zurückgerechnet.
            BOARD_Clock(&dateTime);
            if( dateTime.time.second <= 20 ) seconds -= dateTime.time.second;
            TIMER_Start(&snoozeTimer, seconds);
            DISPLAY_SetBlinkingDigits(0);
            SOUND_Start(SOUND_SNOOZE_BEGIN);
            break;
        }
        case REARM:
        {   SOUND_Start(SOUND_ALARM_REARM);
            DISPLAY_StartAnim(ANIM_ALARM_ENABLE);
            break;
        }
        case ENABLE:
        {   bool alarmEnabled = true;
            STORING_Save(STO_AlarmEnabled, &alarmEnabled);
            SetNextAlarmDateTime(AO_DAILY);
            break;
        }
        case DISABLE:
        {   if( !APPLICATION_RadioManuallyOn() )
            {   BOARD_EnableRelais(false);
                BOARD_SetLedBrightness(0);
            }
            bool alarmEnabled = false;
            STORING_Save(STO_AlarmEnabled, &alarmEnabled);
            break;
        }
        default:
            ;
    }
    if( action != NONE ) APPLICATION_WakeDisplay();
}

//////////////////////////////////////////////////////////////////////////////////////////
