﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Datenaufbewahrung                                                                   //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright © 2010 Reimar Grasbon (softwork@grasbon.de)                               //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "Board.h"
#include "Storing.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define CRC_WEAR_LEVELING_ROTATION   6           // Wechsel zwecks Verschleißreduzierung
#define NO_CRC_MARKER                0xFFFF      // Wertausprägung für "kein CRC"
#define STOP_INDEX                   0xFFFF      // Index Wächterwert

//////////////////////////////////////////////////////////////////////////////////////////
//  Konstanten  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
const Storing factoryDef PROGMEM = // Werksvoreinstellungen
{   .STO_DisplayBlankStart = DEF_DISPLAY_BLANK_START,
    .STO_MotionSensDisplay = DEF_MOTION_SENS_DISPLAY,
    .STO_AlarmTime         = DEF_ALARM_TIME,
    .STO_AlarmEnabled      = DEF_ALARM_ENABLED,
    .STO_AlarmOperation    = DEF_ALARM_OPERATION,
    .STO_AlarmDuration     = DEF_ALARM_DURATION,
    .STO_AlarmTypeRadio    = DEF_ALARM_TYPERADIO,
    .STO_SnoozeEnabled     = DEF_SNOOZE_ENABLED,
    .STO_MotionSensSnooze  = DEF_MOTION_SENS_SNOOZE,
    .STO_SnoozeDuration    = DEF_SNOOZE_DURATION,
    .STO_GermanRegion      = DEF_GERMAN_REGION,
    .STO_ShowHoliday       = DEF_SHOW_HOLIDAY,
    .STO_AutoDstEnabled    = DEF_AUTO_DST_ENABLED
};

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static  UInt16  saveIndex = STOP_INDEX;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////


static void
UpdateCRC(UInt08 byte, UInt16* crc) // 16-Bit CRC nach CCITT
{
    UInt08 i;

    *crc ^= byte;
    for( i=8; i>0; i-- ) *crc = (*crc >> 1) ^ ((*crc&1) ? 0x8408 : 0 );
}

//////////////////////////////////////////////////////////////////////////////////////////
static UInt16
CalcCRC(void)
{
    UInt16  i;
    UInt16  crc;
    UInt08  byte;

    crc = 0xFFFF;
    for( i=0; i<sizeof(Storing); i++ )
    {   BOARD_EEPLoad((const void*)(CFG_EEPROM_STORING_BASE + i), &byte, 1);
        UpdateCRC(byte, &crc);
    }
    if( crc == NO_CRC_MARKER ) crc = ~NO_CRC_MARKER; // Prüfsummenwert ggf. tauschen
    return crc;
}

//////////////////////////////////////////////////////////////////////////////////////////
static UInt16
LoadCRC(void)
{
    const void*  adr;
    UInt16       crc;
    UInt08       i;

    if( CRC_WEAR_LEVELING_ROTATION < 2)
    {   BOARD_EEPLoad((const void*)(CFG_EEPROM_STORING_BASE + sizeof(Storing)),
                      &crc,
                      sizeof(UInt16));
    }
    else
    {   i=0;
        do
        {   adr = (const void*)(CFG_EEPROM_STORING_BASE +
                                sizeof(Storing) + i*sizeof(UInt16));
            BOARD_EEPLoad(adr, &crc, sizeof(UInt16));
            i++;
        } while( i <= CRC_WEAR_LEVELING_ROTATION && crc == NO_CRC_MARKER );
    }
    return crc;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SaveCRC(UInt16 crc)                                    //              EEPMEM:
{                                                      //              +---------------+
    void*   adr;                                       //              |               |
    UInt16  tmpCrc;                                    //              |               |
    UInt08  i;                                         //              |    Storing    |
                                                       //              |     data      |
    if( crc == NO_CRC_MARKER ) crc = ~NO_CRC_MARKER;   //              |               |
    adr = (void*)(CFG_EEPROM_STORING_BASE + sizeof(Storing)); //       |               |
    if( CRC_WEAR_LEVELING_ROTATION < 2 )               //              +---------------+
    {   BOARD_EEPSave(adr, &crc, sizeof(crc));         // CRC       /  | NO_CRC_MARKER | 0
    }                                                  // Wear     |   |       :       | 1
    else                                               // leveling |   |      CRC      | :
    {   // Letzte Prüfsumme suchen                     //          |   |       :       | :
        i=0;                                           //           \  | NO_CRC_MARKER | n
        while( i <= CRC_WEAR_LEVELING_ROTATION )       //              +---------------+
        {   BOARD_EEPLoad(adr, &tmpCrc, sizeof(tmpCrc)); //
            if( tmpCrc != NO_CRC_MARKER ) break;       //   n = CRC_WEAR_LEVELING_ROTATION
            adr += sizeof(UInt16);
            i++;
        }
        if( crc == tmpCrc ) return;
        // Letzte Prüfsumme löschen (NO_CRC_MARKER)
        tmpCrc = NO_CRC_MARKER;
        BOARD_EEPSave(adr, &tmpCrc, sizeof(tmpCrc));
        // Neue Prüfsumme in nachfolgender Zelle ablegen
        i++;
        if( i > CRC_WEAR_LEVELING_ROTATION ) i = 0;
        adr = (void*)(CFG_EEPROM_STORING_BASE + sizeof(Storing) + i*sizeof(crc));
        BOARD_EEPSave(adr, &crc, sizeof(crc));
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Intern  //////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_LoadEEPMem(UInt16 adr, void* mem, UInt16 size)
{
    BOARD_EEPLoad((const void*)CFG_EEPROM_STORING_BASE + adr, mem, size);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_SaveEEPMem(UInt16 adr, const void* mem, UInt16 size)
{
    BOARD_EEPSave((void*)(CFG_EEPROM_STORING_BASE + adr), mem, size);
    saveIndex = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_LoadPGMMem(UInt16 adr, void* mem, UInt16 size)
{
    UInt16  i;

    adr += (UInt16)&factoryDef;
    for( i=0; i<size; i++ ) ((UInt08*)mem)[i] = pgm_read_byte(adr + i);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_RestoreDefaults(void)
{
    UInt16  i;
    UInt08  byte;

    for( i=0; i<sizeof(Storing); i++ )
    {   byte = pgm_read_byte((UInt16)(&factoryDef) + i);
        BOARD_EEPSave((void*)(CFG_EEPROM_STORING_BASE + i), &byte, 1);
    }
    saveIndex = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  System  //////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_Startup(void)
{
    if( CalcCRC() != LoadCRC() ) STORING_RestoreDefaults();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_Proceed(void)
{
    static UInt16  crc = 0;
    UInt08         byte;

    if( saveIndex == STOP_INDEX ) return;
    BOARD_EEPLoad((const void*)(CFG_EEPROM_STORING_BASE + saveIndex), &byte, 1);
    if( saveIndex == 0 ) crc = 0xFFFF;
    UpdateCRC(byte, &crc);
    saveIndex++;
    if( saveIndex == sizeof(Storing) )
    {   saveIndex = STOP_INDEX;
        SaveCRC(crc);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
STORING_Shutdown(void)
{
    while( saveIndex != STOP_INDEX ) STORING_Proceed();
}

//////////////////////////////////////////////////////////////////////////////////////////
